import enum
import mmap
import sys

from xkbcommon._ffi import ffi, lib


class _keepref:
    """Function wrapper that keeps a reference to another object."""
    def __init__(self, ref, func):
        self.ref = ref
        self.func = func

    def __call__(self, *args, **kwargs):
        self.func(*args, **kwargs)


# enum.IntFlag was changed in a non-backward-compatible manner in
# Python 3.11: the __str__ method was changed to int.__str__(). We
# retain the old behaviour because it is useful, and in order not to
# surprise existing users of python-xkbcommon

if sys.version_info < (3, 11):
    _IntFlag = enum.IntFlag
else:
    class _IntFlag(int, enum.Flag, boundary=enum.KEEP):
        """IntFlag implementation that retains the pre-python-3.11 behaviour
        """
        pass


class XKBError(Exception):
    """Base for all XKB exceptions"""
    pass


class XKBPathError(Exception):
    """There was a problem altering the include path of a Context."""
    pass


class XKBBufferTooSmall(XKBError):
    """A buffer created for libxkbcommon to return data in was too small.

    This is a limit in python-xkbcommon.  If you run into it, please
    file a bug report and the limit can be raised.
    """
    pass


class XKBInvalidKeysym(XKBError):
    pass


class XKBKeymapCreationFailure(XKBError):
    """Unable to create a keymap."""
    pass


class XKBKeymapReadError(XKBError):
    """Unable to fetch a keymap as a string."""
    pass


class XKBInvalidModifierIndex(XKBError):
    pass


class XKBInvalidKeycode(XKBError):
    pass


class XKBKeyDoesNotExist(XKBError):
    """A given key name does not exist.
    """
    def __init__(self, key_name):
        super().__init__(key_name)
        self.key_name = key_name


class XKBModifierDoesNotExist(XKBError):
    """A given modifier name does not exist.

    If the name is one of a list of names and libxkbcommon does not
    indicate which of the list does not exist, modifier_name will be
    None.
    """
    def __init__(self, modifier_name):
        super().__init__(modifier_name)
        self.modifier_name = modifier_name


class XKBInvalidLayoutIndex(XKBError):
    pass


class XKBLayoutDoesNotExist(XKBError):
    def __init__(self, index_name):
        super().__init__(index_name)
        self.index_name = index_name


class XKBInvalidLEDIndex(XKBError):
    pass


class XKBLEDDoesNotExist(XKBError):
    def __init__(self, led_name):
        super().__init__(led_name)
        self.led_name = led_name


class XKBComposeTableCreationFailure(XKBError):
    """Unable to create a compose table."""
    pass


class XKBComposeStateCreationFailure(XKBError):
    """Unable to create a compose state."""
    pass


# Internal helper for logging callback
def _onerror_do_nothing(exception, exc_value, traceback):
    return


@ffi.def_extern(onerror=_onerror_do_nothing)
def _log_handler(user_data, level, message):
    context = ffi.from_handle(user_data)
    if context._log_fn:
        context._log_fn(context, level, ffi.string(message).decode('utf8'))


# Internal helper for keycode iteration
@ffi.def_extern(onerror=_onerror_do_nothing)
def _key_for_each_helper(keymap, key, data):
    k = ffi.from_handle(data)
    k.append(key)


# Keysyms http://xkbcommon.org/doc/current/group__keysyms.html

def keysym_get_name(keysym):
    "Get the name of a keysym."
    name = ffi.new("char[64]")
    r = lib.xkb_keysym_get_name(keysym, name, len(name))
    if r == -1:
        raise XKBInvalidKeysym()
    if r > len(name):
        raise XKBBufferTooSmall()
    return ffi.string(name).decode('ascii')


def keysym_from_name(name, case_insensitive=False):
    "Get a keysym from its name."
    flags = 0
    if case_insensitive:
        flags = flags | lib.XKB_KEYSYM_CASE_INSENSITIVE
    return lib.xkb_keysym_from_name(name.encode('ascii'), flags)


def keysym_to_string(keysym):
    buffer = ffi.new("char[64]")
    r = lib.xkb_keysym_to_utf8(keysym, buffer, len(buffer))
    if r == -1:
        raise XKBBufferTooSmall()
    if r == 0:
        return
    return ffi.string(buffer).decode('utf8')


def keysym_to_upper(keysym):
    """Convert a keysym to its uppercase form.

    If there is no such form, the keysym is returned unchanged.

    The conversion rules may be incomplete; prefer to work with the
    Unicode representation instead, when possible.
    """
    return lib.xkb_keysym_to_upper(keysym)


def keysym_to_lower(keysym):
    """Convert a keysym to its lowercase form.

    The conversion rules may be incomplete; prefer to work with the
    Unicode representation instead, when possible.
    """
    return lib.xkb_keysym_to_lower(keysym)


# Compose enumerations required by Context object

@enum.unique
class ComposeCompileFlags(_IntFlag):
    """Flags affecting Compose file compilation
    """
    pass


@enum.unique
class ComposeFormat(enum.IntEnum):
    """The recognized Compose file formats

    XKB_COMPOSE_FORMAT_TEXT_V1: The classic libX11 Compose text format
    """
    XKB_COMPOSE_FORMAT_TEXT_V1 = lib.XKB_COMPOSE_FORMAT_TEXT_V1


# Library Context http://xkbcommon.org/doc/current/group__context.html

class Context:
    """xkbcommon library context.

    Every keymap compilation request must have a context associated
    with it. The context keeps around state such as the include path.

    The user may set some environment variables which affect the
    library:

    XKB_CONFIG_ROOT, HOME - affect include paths
    XKB_LOG_LEVEL - see set_log_level().
    XKB_LOG_VERBOSITY - see set_log_verbosity().
    XKB_DEFAULT_RULES
    XKB_DEFAULT_MODEL
    XKB_DEFAULT_LAYOUT
    XKB_DEFAULT_VARIANT
    XKB_DEFAULT_OPTIONS - see xkb_rule_names.
    """

    def __init__(self, no_default_includes=False, no_environment_names=False,
                 no_secure_getenv=False):
        """Create a new context.

        Keyword arguments:

        no_default_includes: if set, create this context with an empty
        include path.

        no_environment_names: if set, don't take RMLVO names from the
        environment.

        no_secure_getenv: if set, use getenv() instead of
        secure_getenv() to obtain environment variables.
        """
        flags = lib.XKB_CONTEXT_NO_FLAGS
        if no_default_includes:
            flags = flags | lib.XKB_CONTEXT_NO_DEFAULT_INCLUDES
        if no_environment_names:
            flags = flags | lib.XKB_CONTEXT_NO_ENVIRONMENT_NAMES
        if no_secure_getenv:
            flags = flags | lib.XKB_CONTEXT_NO_SECURE_GETENV
        context = lib.xkb_context_new(flags)
        if not context:
            raise XKBError("Couldn't create XKB context")
        self._context = ffi.gc(context, _keepref(lib, lib.xkb_context_unref))
        self._log_fn = None
        # We keep a reference to the handle to keep it alive
        self._userdata = ffi.new_handle(self)
        lib.xkb_context_set_user_data(self._context, self._userdata)

    # Include Paths http://xkbcommon.org/doc/current/group__include-path.html

    def include_path_append(self, path):
        "Append a new entry to the context's include path."
        r = lib.xkb_context_include_path_append(
            self._context, path.encode('utf8'))
        if r != 1:
            raise XKBPathError("Failed to append to include path")

    def include_path_append_default(self):
        "Append the default include paths to the context's include path."
        r = lib.xkb_context_include_path_append_default(self._context)
        if r != 1:
            raise XKBPathError("Failed to append default include paths")

    def include_path_reset_defaults(self):
        """Reset the context's include path to the default.

        Removes all entries from the context's include path, and
        inserts the default paths.
        """
        r = lib.xkb_context_include_path_reset_defaults(self._context)
        if r != 1:
            raise XKBPathError("Failed to restore default include path")

    def include_path_get(self, index):
        """Get a specific include path from the context's include path.

        Returns the include path at the specified index.  If the index
        is invalid, returns None.
        """
        assert isinstance(index, int)
        r = lib.xkb_context_include_path_get(self._context, index)
        if r:
            return ffi.string(r).decode("utf8")

    def num_include_paths(self):
        """Return the number of include paths"""
        return lib.xkb_context_num_include_paths(self._context)

    def include_path(self):
        "Iterate over the include path."
        i = 0
        while True:
            p = self.include_path_get(i)
            if not p:
                return
            yield p
            i += 1

    # Logging Handling http://xkbcommon.org/doc/current/group__logging.html

    def set_log_level(self, level):
        """Set the current logging level.

        The default level is XKB_LOG_LEVEL_ERROR. The environment
        variable XKB_LOG_LEVEL, if set at the time the context was
        created, overrides the default value.  It may be specified as
        a level number or name.
        """
        lib.xkb_context_set_log_level(self._context, level)

    def get_log_level(self):
        """Return the current logging level."""
        return lib.xkb_context_get_log_level(self._context)

    def set_log_verbosity(self, verbosity):
        """Set the current logging verbosity.

        The library can generate a number of warnings which are not
        helpful to ordinary users of the library. The verbosity may be
        increased if more information is desired (e.g. when developing
        a new keymap).

        The default verbosity is 0. The environment variable
        XKB_LOG_VERBOSITY, if set at the time the context was created,
        overrides the default value.

        Currently used values are 1 to 10, higher values being more
        verbose.  0 would result in no verbose messages being logged.

        Most verbose messages are of level XKB_LOG_LEVEL_WARNING or lower.
        """
        lib.xkb_context_set_log_verbosity(self._context, verbosity)

    def get_log_verbosity(self):
        """Return the current logging verbosity."""
        return lib.xkb_context_get_log_verbosity(self._context)

    def set_log_fn(self, handler):
        """Set a custom function to handle logging messages.

        By default, log messages from this library are printed to
        stderr. This function allows you to replace the default
        behavior with a custom handler. The handler is only called
        with messages which match the current logging level and
        verbosity settings for the context.

        The handler is called with the following arguments:
        - context: this object
        - level: the logging level of the message
        - message: the message itself

        Passing None as the handler restores the default function,
        which logs to stderr.

        NB this implementation uses xkb_context_set_user_data() on the
        context.  Don't call xkb_context_set_user_data() if you intend
        to install a custom function to handle logging messages.
        """
        if handler:
            lib._set_log_handler_internal(self._context)
            self._log_fn = handler
        else:
            lib.xkb_context_set_log_fn(self._context, ffi.NULL)

    # Keymap Creation http://xkbcommon.org/doc/current/group__keymap.html

    def keymap_new_from_names(self, rules=None, model=None, layout=None,
                              variant=None, options=None):
        """Create a keymap from RMLVO names.

        The primary keymap entry point: creates a new XKB keymap from
        a set of RMLVO (Rules + Model + Layouts + Variants + Options)
        names.

        Returns a Keymap compiled according to the RMLVO names, or
        None if the compilation failed.
        """
        # CFFI memory management note:
        # The C strings allocated below using ffi.new("char[]", ...)
        # are automatically deallocated when there are no remaining
        # python references to them. Being assigned to members of the
        # 'names' struct does not count as a reference! We keep
        # references to them in the 'keep_alive' list until after the
        # call to xkb_keymap_new_from_names() to avoid problems with
        # use-after-free.
        names = ffi.new("struct xkb_rule_names *")
        keep_alive = []
        for x in ("rules", "model", "layout", "variant", "options"):
            if locals()[x]:
                c = ffi.new("char[]", locals()[x].encode())
                setattr(names, x, c)
                keep_alive.append(c)
        r = lib.xkb_keymap_new_from_names(
            self._context, names, lib.XKB_KEYMAP_COMPILE_NO_FLAGS)
        del keep_alive, names
        if r == ffi.NULL:
            raise XKBKeymapCreationFailure(
                "xkb_keymap_new_from_names returned NULL")
        return Keymap(self, r, "names")

    # We can't call xkb_keymap_new_from_file directly because we have
    # no way to get a C stdio FILE * for a file opened by python.  We
    # can fake it by mmaping the file and using
    # xkb_keymap_new_from_{string,buffer} instead - which happens to
    # be what libxkbcommon does internally anyway!
    def keymap_new_from_file(self, file, format=lib.XKB_KEYMAP_FORMAT_TEXT_V1):
        "Create a Keymap from an open file"
        try:
            fn = file.fileno()
        except Exception:
            load_method = "read_file"
            keymap = file.read()
            r = lib.xkb_keymap_new_from_string(
                self._context, keymap, format,
                lib.XKB_KEYMAP_COMPILE_NO_FLAGS)
        else:
            load_method = "mmap_file"
            mm = mmap.mmap(fn, 0)
            buf = ffi.from_buffer(mm)
            r = lib.xkb_keymap_new_from_buffer(
                self._context, buf, mm.size(), format,
                lib.XKB_KEYMAP_COMPILE_NO_FLAGS)
            del buf
            mm.close()

        if r == ffi.NULL:
            raise XKBKeymapCreationFailure(
                "xkb_keymap_new_from_buffer or xkb_keymap_new_from_string "
                "returned NULL")
        return Keymap(self, r, load_method)

    def keymap_new_from_string(
            self, string, format=lib.XKB_KEYMAP_FORMAT_TEXT_V1):
        "Create a Keymap from a keymap string."
        r = lib.xkb_keymap_new_from_string(
            self._context, string.encode("ascii"), format,
            lib.XKB_KEYMAP_COMPILE_NO_FLAGS)
        if r == ffi.NULL:
            raise XKBKeymapCreationFailure(
                "xkb_keymap_new_from_string returned NULL")
        return Keymap(self, r, "string")

    def keymap_new_from_buffer(
            self, buffer, format=lib.XKB_KEYMAP_FORMAT_TEXT_V1, length=None):
        "Create a Keymap from a memory buffer."
        buf = ffi.from_buffer(buffer)
        r = lib.xkb_keymap_new_from_buffer(
            self._context, buf, length if length else len(buf), format,
            lib.XKB_KEYMAP_COMPILE_NO_FLAGS)
        if r == ffi.NULL:
            raise XKBKeymapCreationFailure(
                "xkb_keymap_new_from_buffer returned NULL")
        return Keymap(self, r, "buffer")

    # Compose table creation
    # http://xkbcommon.org/doc/current/group__compose.html

    def compose_table_new_from_locale(self, locale, flags=None):
        """Create a compose table for a given locale.

        The locale is used for searching the file-system for an
        appropriate Compose file. The search order is described in
        Compose(5). It is affected by the following environment
        variables:

        XCOMPOSEFILE - see Compose(5).

        XDG_CONFIG_HOME - before $HOME/.XCompose is checked,
        $XDG_CONFIG_HOME/XCompose is checked (with a fall back to
        $HOME/.config/XCompose if XDG_CONFIG_HOME is not
        defined). This is a libxkbcommon extension to the search
        procedure in Compose(5) (since libxkbcommon 1.0.0). Note that
        other implementations, such as libX11, might not find a
        Compose file in this path.

        HOME - see Compose(5).

        XLOCALEDIR - if set, used as the base directory for the
        system's X locale files, e.g. /usr/share/X11/locale, instead
        of the preconfigured directory.
        """
        c_locale = ffi.new("char[]", locale.encode())
        table = lib.xkb_compose_table_new_from_locale(
            self._context, c_locale, flags if flags else 0)
        if not table:
            raise XKBComposeTableCreationFailure(
                f"Could not create compose table from locale '{locale}'")
        return ComposeTable(self, table, "locale")

    # See keymap_new_from_file() for note about FILE *
    def compose_table_new_from_file(
            self, file, locale,
            format=ComposeFormat.XKB_COMPOSE_FORMAT_TEXT_V1, flags=None):
        "Create a ComposeTable from an open file"
        try:
            fn = file.fileno()
        except Exception:
            data = file.read()
            return self._compose_table_new_from_buffer_internal(
                data, locale, "read_file", format, flags)
        else:
            with mmap.mmap(fn, 0) as mm:
                return self._compose_table_new_from_buffer_internal(
                    mm, locale, "mmap_file", format, flags)

    def compose_table_new_from_buffer(
            self, buffer, locale,
            format=ComposeFormat.XKB_COMPOSE_FORMAT_TEXT_V1, flags=None,
            length=None):
        "Create a ComposeTable from a memory buffer."
        return self._compose_table_new_from_buffer_internal(
            buffer, locale, "buffer", format, flags, length)

    def _compose_table_new_from_buffer_internal(
            self, buffer, locale, load_type,
            format=ComposeFormat.XKB_COMPOSE_FORMAT_TEXT_V1, flags=None,
            length=None):
        c_locale = ffi.new("char[]", locale.encode())
        buf = ffi.from_buffer(buffer)
        table = lib.xkb_compose_table_new_from_buffer(
            self._context, buf, length if length else len(buf), c_locale,
            format, flags if flags else 0)
        del buf
        if not table:
            raise XKBComposeTableCreationFailure(
                "xkb_compose_table_new_from_buffer returned NULL")
        return ComposeTable(self, table, load_type)


class Keymap:
    """A keymap.

    Do not instantiate this object directly.  Instead, use the various
    'keymap_new_from_' methods of Context.
    """

    def __init__(self, context, pointer, load_method):
        self.load_method = load_method
        self._context = context

        self._keymap = ffi.gc(pointer, _keepref(lib, lib.xkb_keymap_unref))
        self._valid_keycodes = None

    def get_as_bytes(self, format=lib.XKB_KEYMAP_FORMAT_TEXT_V1):
        """Get the compiled keymap as bytes.

        On Python 3 will return a bytes object.

        On Python 2 will return a str object.
        """
        r = lib.xkb_keymap_get_as_string(self._keymap, format)
        if r == ffi.NULL:
            raise XKBKeymapReadError()
        kms = ffi.string(r)
        lib.free(r)
        return kms

    def get_as_string(self, format=lib.XKB_KEYMAP_FORMAT_TEXT_V1):
        """Get the compiled keymap as a string.

        On Python 3 will return a str object.

        On Python 2 will return a unicode object.
        """
        return self.get_as_bytes(format).decode('ascii')

    # Keymap Components http://xkbcommon.org/doc/current/group__components.html

    def min_keycode(self):
        "Get the minimum keycode in the keymap."
        return lib.xkb_keymap_min_keycode(self._keymap)

    def max_keycode(self):
        "Get the maximum keycode in the keymap."
        return lib.xkb_keymap_max_keycode(self._keymap)

    # The xkb_keymap_key_for_each() call isn't very amenable to being
    # used directly from python.  It's much more useful to implement a
    # python iterable instead.
    def _get_valid_keycodes(self):
        """Fetch a list of valid keycodes in the keymap.

        Uses the xkb_keymap_key_for_each() call.
        """
        keycodes = []
        keycode_ref = ffi.new_handle(keycodes)
        lib.xkb_keymap_key_for_each(
            self._keymap, lib._key_for_each_helper, keycode_ref)
        self._valid_keycodes = keycodes

    def __iter__(self):
        """Iterate over valid keycodes"""
        if self._valid_keycodes is None:
            self._get_valid_keycodes()
        return iter(self._valid_keycodes)

    def key_get_name(self, key):
        r = lib.xkb_keymap_key_get_name(self._keymap, key)
        if r == ffi.NULL:
            raise XKBInvalidKeycode
        return ffi.string(r).decode('ascii')

    def key_by_name(self, name):
        r = lib.xkb_keymap_key_by_name(self._keymap, name.encode('ascii'))
        if r == lib.XKB_KEYCODE_INVALID:
            raise XKBKeyDoesNotExist(name)
        return r

    def num_mods(self):
        """Get the number of modifiers in the keymap."""
        return lib.xkb_keymap_num_mods(self._keymap)

    def mod_get_name(self, idx):
        """Get the name of a modifier by index.

        Returns the name.  If the index is invalid, raises
        XKBInvalidModifierIndex.
        """
        r = lib.xkb_keymap_mod_get_name(self._keymap, idx)
        if r == ffi.NULL:
            raise XKBInvalidModifierIndex()
        return ffi.string(r).decode('ascii')

    def mod_get_index(self, name):
        """Get the index of a modifier by name.

        Returns the index.  If no modifier with this name exists,
        raises XKBModifierDoesNotExist.
        """
        r = lib.xkb_keymap_mod_get_index(self._keymap, name.encode('ascii'))
        if r == lib.XKB_MOD_INVALID:
            raise XKBModifierDoesNotExist(name)
        return r

    def num_layouts(self):
        """Get the number of layouts in the keymap."""
        return lib.xkb_keymap_num_layouts(self._keymap)

    def layout_get_name(self, idx):
        """Get the name of a layout by index.

        Returns the name.  If the layout does not have a name, returns
        None.  If the index is invalid, raises XKBInvalidLayoutIndex.
        """
        if idx >= self.num_layouts():
            raise XKBInvalidLayoutIndex()
        r = lib.xkb_keymap_layout_get_name(self._keymap, idx)
        if r == ffi.NULL:
            return
        return ffi.string(r).decode('ascii')

    def layout_get_index(self, name):
        """Get the index of a layout by name.

        Returns the index.  If no layout exists with this name, raises
        XKBLayoutDoesNotExist. If more than one layout in the keymap
        has this name, returns the lowest index among them.
        """
        r = lib.xkb_keymap_layout_get_index(self._keymap, name.encode('ascii'))
        if r == lib.XKB_LAYOUT_INVALID:
            raise XKBLayoutDoesNotExist(name)
        return r

    def num_leds(self):
        """Get the number of LEDs in the keymap.

        Warning: The range [ 0..Keymap.num_leds() ) includes all of
        the LEDs in the keymap, but may also contain inactive
        LEDs. When iterating over this range, you need to handle this
        case when calling functions such as Keymap.led_get_name()
        or State.led_index_is_active().
        """
        return lib.xkb_keymap_num_leds(self._keymap)

    def led_get_name(self, idx):
        """Get the name of a LED by index.

        Returns the name. If the index is invalid, returns None.
        """
        r = lib.xkb_keymap_led_get_name(self._keymap, idx)
        if r == ffi.NULL:
            raise XKBInvalidLEDIndex()
        return ffi.string(r).decode('ascii')

    def led_get_index(self, name):
        """Get the index of a LED by name.

        Returns the index. If no LED with this name exists, returns
        lib.XKB_LED_INVALID.
        """
        r = lib.xkb_keymap_led_get_index(self._keymap, name.encode('ascii'))
        if r == lib.XKB_LED_INVALID:
            raise XKBLEDDoesNotExist(name)
        return r

    def num_layouts_for_key(self, key):
        """Get the number of layouts for a specific key.

        This number can be different from Keymap.num_layouts(), but is
        always smaller. It is the appropriate value to use when
        iterating over the layouts of a key.
        """
        return lib.xkb_keymap_num_layouts_for_key(self._keymap, key)

    def num_levels_for_key(self, key, layout):
        """Get the number of shift levels for a specific key and layout.

        If layout is out of range for this key (that is, larger or
        equal to the value returned by Keymap.num_layouts_for_key()),
        it is brought back into range in a manner consistent with
        State.key_get_layout().
        """
        return lib.xkb_keymap_num_levels_for_key(self._keymap, key, layout)

    def key_get_mods_for_level(self, key, layout, level):
        """Retrieves every possible modifier mask that produces the specified
        shift level for a specific key and layout.

        This API is useful for inverse key transformation;
        i.e. finding out which modifiers need to be active in order to
        be able to type the keysym(s) corresponding to the specific
        key code, layout and level.

        If layout is out of range for this key (that is, larger or
        equal to the value returned by Keymap.num_layouts_for_key()),
        it is brought back into range in a manner consistent with
        State.key_get_layout().
        """
        masks_size = 4
        while True:
            masks_out = ffi.new(f"xkb_mod_mask_t[{masks_size}]")
            r = lib.xkb_keymap_key_get_mods_for_level(
                self._keymap, key, layout, level, masks_out, masks_size)
            if r < masks_size:
                return [masks_out[n] for n in range(r)]
            masks_size = masks_size << 1

    def key_get_syms_by_level(self, key, layout, level):
        """Get the keysyms obtained from pressing a key in a given layout and
        shift level.

        This function is like State.key_get_syms(), only the layout
        and shift level are not derived from the keyboard state but
        are instead specified explicitly.

        Returns a list of keysyms.
        """
        syms_out = ffi.new("const xkb_keysym_t **")
        r = lib.xkb_keymap_key_get_syms_by_level(
            self._keymap, key, layout, level, syms_out)
        syms = []
        if r > 0:
            assert syms_out[0] != ffi.NULL
        for i in range(0, r):
            syms.append(syms_out[0][i])
        return syms

    def key_repeats(self, key):
        """Determine whether a key should repeat or not.

        A keymap may specify different repeat behaviors for different
        keys. Most keys should generally exhibit repeat behavior; for
        example, holding the 'a' key down in a text editor should
        normally insert a single 'a' character every few milliseconds,
        until the key is released. However, there are keys which
        should not or do not need to be repeated. For example,
        repeating modifier keys such as Left/Right Shift or Caps Lock
        is not generally useful or desired.

        Returns True if the key should repeat, False otherwise.
        """
        return lib.xkb_keymap_key_repeats(self._keymap, key) == 1

    # Keyboard State http://xkbcommon.org/doc/current/group__state.html

    def state_new(self):
        """Create a new keyboard state object.

        Returns a new keyboard state object.  Raises XKBError on
        failure.
        """
        return KeyboardState(self)


@enum.unique
class KeyDirection(enum.IntEnum):
    "Specifies the direction of the key (press / release)"
    XKB_KEY_UP = lib.XKB_KEY_UP
    XKB_KEY_DOWN = lib.XKB_KEY_DOWN


@enum.unique
class StateComponent(_IntFlag):
    """Modifier and layout types for state objects

    This enum is bitmaskable, e.g. (XKB_STATE_MODS_DEPRESSED |
    XKB_STATE_MODS_LATCHED) is valid to exclude locked modifiers.

    In XKB, the DEPRESSED components are also known as 'base'.
    """
    XKB_STATE_MODS_DEPRESSED = lib.XKB_STATE_MODS_DEPRESSED
    XKB_STATE_MODS_LATCHED = lib.XKB_STATE_MODS_LATCHED
    XKB_STATE_MODS_LOCKED = lib.XKB_STATE_MODS_LOCKED
    XKB_STATE_MODS_EFFECTIVE = lib.XKB_STATE_MODS_EFFECTIVE
    XKB_STATE_LAYOUT_DEPRESSED = lib.XKB_STATE_LAYOUT_DEPRESSED
    XKB_STATE_LAYOUT_LATCHED = lib.XKB_STATE_LAYOUT_LATCHED
    XKB_STATE_LAYOUT_LOCKED = lib.XKB_STATE_LAYOUT_LOCKED
    XKB_STATE_LAYOUT_EFFECTIVE = lib.XKB_STATE_LAYOUT_EFFECTIVE
    XKB_STATE_LEDS = lib.XKB_STATE_LEDS


@enum.unique
class StateMatch(_IntFlag):
    """State match flags

    Match flags for KeyboardState.mod_indices_are_active() and
    KeyboardState.mod_names_are_active(), specifying the conditions
    for a successful match.

    XKB_STATE_MATCH_NON_EXCLUSIVE is bitmaskable with the other modes.
    """
    XKB_STATE_MATCH_ANY = lib.XKB_STATE_MATCH_ANY
    XKB_STATE_MATCH_ALL = lib.XKB_STATE_MATCH_ALL
    XKB_STATE_MATCH_NON_EXCLUSIVE = lib.XKB_STATE_MATCH_NON_EXCLUSIVE


@enum.unique
class ConsumedMode(enum.IntEnum):
    """Consumed modifiers mode

    There are several possible methods for deciding which modifiers
    are consumed and which are not, each applicable for different
    systems or situations. The mode selects the method to use.

    Keep in mind that in all methods, the keymap may decide to
    "preserve" a modifier, meaning it is not reported as consumed even
    if it would have otherwise.
    """
    XKB_CONSUMED_MODE_XKB = lib.XKB_CONSUMED_MODE_XKB
    XKB_CONSUMED_MODE_GTK = lib.XKB_CONSUMED_MODE_GTK


@enum.unique
class ComposeStateFlags(_IntFlag):
    """Flags for compose state creation
    """
    pass


@enum.unique
class ComposeStatus(enum.IntEnum):
    """Status of the Compose sequence state machine

    XKB_COMPOSE_NOTHING: The initial state; no sequence has started yet
    XKB_COMPOSE_COMPOSING: In the middle of a sequence
    XKB_COMPOSE_COMPOSED: A complete sequence has been matched
    XKB_COMPOSE_CANCELLED: The last sequence was cancelled due to an
      unmatched keysym
    """
    XKB_COMPOSE_NOTHING = lib.XKB_COMPOSE_NOTHING
    XKB_COMPOSE_COMPOSING = lib.XKB_COMPOSE_COMPOSING
    XKB_COMPOSE_COMPOSED = lib.XKB_COMPOSE_COMPOSED
    XKB_COMPOSE_CANCELLED = lib.XKB_COMPOSE_CANCELLED


@enum.unique
class ComposeFeedResult(enum.IntEnum):
    """The effect of a keysym fed to ComposeState.feed()

    XKB_COMPOSE_FEED_IGNORED: The keysym had no effect
    XKB_COMPOSE_FEED_ACCEPTED: The keysym started, advanced or cancelled
      a sequence
    """
    XKB_COMPOSE_FEED_IGNORED = lib.XKB_COMPOSE_FEED_IGNORED
    XKB_COMPOSE_FEED_ACCEPTED = lib.XKB_COMPOSE_FEED_ACCEPTED


# Global names for enum members
for _ec in (KeyDirection, StateComponent, StateMatch, ConsumedMode,
            ComposeCompileFlags, ComposeFormat, ComposeStateFlags,
            ComposeStatus, ComposeFeedResult):
    for _sc in _ec:
        globals()[_sc.name] = _sc


class KeyboardState:
    def __init__(self, keymap):
        state = lib.xkb_state_new(keymap._keymap)
        if not state:
            raise XKBError("Couldn't create keyboard state")
        # Keep the keymap around to ensure it isn't collected too soon
        self.keymap = keymap
        self._state = ffi.gc(state, _keepref(lib, lib.xkb_state_unref))

    def get_keymap(self):
        """Get the Keymap which a keyboard state object is using.

        The keymap is also accessible as the "keymap" attribute.
        """
        return self.keymap

    def update_key(self, key, direction):
        """Update the keyboard state to reflect a given key being pressed or
        released.

        This entry point is intended for programs which track the
        keyboard state explicitly (like an evdev client). If the state
        is serialized to you by a master process (like a Wayland
        compositor) using functions like KeyboardState.serialize_mods(),
        you should use KeyboardState.update_mask() instead. The two
        functions should not generally be used together.

        A series of calls to this function should be consistent; that
        is, a call with XKB_KEY_DOWN for a key should be matched by an
        XKB_KEY_UP; if a key is pressed twice, it should be released
        twice; etc. Otherwise (e.g. due to missed input events),
        situations like "stuck modifiers" may occur.

        This function is often used in conjunction with the function
        KeyboardState.key_get_syms() (or
        KeyboardState.key_get_one_sym()), for example, when handling a
        key event. In this case, you should prefer to get the keysyms
        before updating the key, such that the keysyms reported for
        the key event are not affected by the event itself. This is
        the conventional behavior.

        Returns a mask of state components that have changed as a
        result of the update. If nothing in the state has changed,
        returns 0.
        """
        return StateComponent(
            lib.xkb_state_update_key(self._state, key, direction))

    def update_mask(self, depressed_mods, latched_mods, locked_mods,
                    depressed_layout, latched_layout, locked_layout):
        """Update a keyboard state from a set of explicit masks.

        This entry point is intended for window systems and the like,
        where a master process holds an xkb_state, then serializes it
        over a wire protocol, and clients then use the serialization
        to feed in to their own xkb_state.

        All parameters must always be passed, or the resulting state
        may be incoherent.

        The serialization is lossy and will not survive round trips;
        it must only be used to feed slave state objects, and must not
        be used to update the master state.

        If you do not fit the description above, you should use
        KeyboardState.update_key() instead. The two functions should
        not generally be used together.

        Returns a mask of state components that have changed as a
        result of the update. If nothing in the state has changed,
        returns 0.
        """
        return StateComponent(lib.xkb_state_update_mask(
            self._state, depressed_mods, latched_mods, locked_mods,
            depressed_layout, latched_layout, locked_layout))

    def key_get_syms(self, key):
        """Get the keysyms obtained from pressing a particular key in a given
        keyboard state.

        Get the keysyms for a key according to the current active
        layout, modifiers and shift level for the key, as determined
        by a keyboard state.

        As an extension to XKB, this function can return more than one
        keysym. If you do not want to handle this case, you can use
        KeyboardState.key_get_one_sym() for a simpler interface.

        This function does not perform any Keysym
        Transformations. (This might change).
        """
        syms_out = ffi.new("const xkb_keysym_t **")
        r = lib.xkb_state_key_get_syms(self._state, key, syms_out)
        syms = []
        if r > 0:
            assert syms_out[0] != ffi.NULL
        for i in range(0, r):
            syms.append(syms_out[0][i])
        return syms

    def key_get_string(self, key):
        """Get the Unicode/UTF-8 string obtained from pressing a particular
        key in a given keyboard state.

        This function performs Capitalization and Control Keysym
        Transformations.

        Returns the string.  If there is nothing to return, returns
        the empty string.
        """
        buffer = ffi.new("char[64]")
        r = lib.xkb_state_key_get_utf8(self._state, key, buffer, len(buffer))
        if r + 1 > len(buffer):
            raise XKBBufferTooSmall()
        return ffi.string(buffer).decode('utf8')

    def key_get_one_sym(self, keycode):
        """Get the single keysym obtained from pressing a particular key in a
        given keyboard state.

        This function is similar to xkb_state_key_get_syms(), but
        intended for users which cannot or do not want to handle the
        case where multiple keysyms are returned (in which case this
        function is preferred).

        This function performs Capitalization Keysym Transformations.

        Returns the keysym. If the key does not have exactly one
        keysym, returns lib.XKB_KEY_NoSymbol.
        """
        return lib.xkb_state_key_get_one_sym(self._state, keycode)

    def key_get_layout(self, key):
        """Get the effective layout index for a key in a given keyboard state.

        Returns the layout index for the key in the given keyboard
        state. If the given keycode is invalid, or if the key is not
        included in any layout at all, raises XKBInvalidLayoutIndex.
        """
        r = lib.xkb_state_key_get_layout(self._state, key)
        if r == lib.XKB_LAYOUT_INVALID:
            raise XKBInvalidLayoutIndex()
        return r

    def key_get_level(self, key, layout):
        """Get the effective shift level for a key in a given keyboard state
        and layout.

        layout must be smaller than:
        State.get_keymap().num_layouts_for_key(key)

        Usually layout would be:
        State.key_get_layout(key)

        Returns the shift level index. If the key or layout are
        invalid, raises XKBInvalidLayoutIndex.
        """
        r = lib.xkb_state_key_get_level(self._state, key, layout)
        if r == lib.XKB_LEVEL_INVALID:
            raise XKBInvalidLayoutIndex()
        return r

    def serialize_mods(self, components):
        """The counterpart to xkb_state_update_mask for modifiers, to be used
        on the server side of serialization.

        components is a mask of the modifier state components to
        serialize. State components other than XKB_STATE_MODS_* are
        ignored. If XKB_STATE_MODS_EFFECTIVE is included, all other
        state components are ignored.

        Returns a xkb_mod_mask_t representing the given components of
        the modifier state.

        This function should not be used in regular clients; please
        use the State.mod_*_is_active() API instead.
        """
        return lib.xkb_state_serialize_mods(self._state, components)

    def serialize_layout(self, components):
        """The counterpart to xkb_state_update_mask for layouts, to be used on
        the server side of serialization.

        components is a mask of the modifier state components to
        serialize. State components other than XKB_STATE_MODS_* are
        ignored. If XKB_STATE_MODS_EFFECTIVE is included, all other
        state components are ignored.

        Returns a layout index representing the given components of
        the layout state.

        This function should not be used in regular clients; please
        use the State.layout_*_is_active() API instead.
        """
        return lib.xkb_state_serialize_layout(self._state, components)

    def mod_name_is_active(self, name, type):
        """Test whether a modifier is active in a given keyboard state by
        name.

        type is the component of the state against which to match the
        given modifiers.

        Returns True if the modifier is active, False if it is not.
        If the modifier name does not exist in the keymap, raises
        XKBModifierDoesNotExist.
        """
        r = lib.xkb_state_mod_name_is_active(self._state, name.encode(), type)
        if r == -1:
            raise XKBModifierDoesNotExist(name)
        return r == 1

    def mod_names_are_active(self, type, match, names):
        """Test whether a set of modifiers are active in a given keyboard
        state by name.

        type is the component of the state against which to match the
        given modifiers.

        match is the manner by which to match the state against the
        given modifiers.

        names is an iterable of modifier names to test.

        Returns True if the modifiers are active, False if they are
        not.  If any of the modifier names do not exist, raises
        XKBModifierDoesNotExist(None).
        """
        args = [ffi.new("char []", n.encode('ascii')) for n in names]
        args.append(ffi.NULL)
        r = lib.xkb_state_mod_names_are_active(self._state, type, match, *args)
        if r == -1:
            raise XKBModifierDoesNotExist(None)
        return r == 1

    def mod_index_is_active(self, idx, type):
        """Test whether a modifier is active in a given keyboard state by
        index.

        Returns True if the modifier is active, False if it is not.
        If the modifier index is invalid in the keymap, raises
        XKBInvalidModifierIndex.
        """
        r = lib.xkb_state_mod_index_is_active(self._state, idx, type)
        if r == -1:
            raise XKBInvalidModifierIndex()
        return r == 1

    def mod_indices_are_active(self, type, match, mods):
        """Test whether a set of modifiers are active in a given keyboard
        state by index.

        type is the component of the state against which to match the
        given modifiers.

        match is the manner by which to match the state against the
        given modifiers.

        mods is an iterable of modifier indices to test.

        Returns True if the modifiers are active, False if they are
        not.  If any of the modifier indices are invalid in the
        keymap, raises XKBInvalidModifierIndex.
        """
        args = [ffi.cast("int", x) for x in mods]
        args.append(ffi.cast("int", lib.XKB_MOD_INVALID))
        r = lib.xkb_state_mod_indices_are_active(
            self._state, type, match, *args)
        if r == -1:
            raise XKBInvalidModifierIndex()
        return r == 1

    def mod_index_is_consumed(self, key, idx,
                              mode=ConsumedMode.XKB_CONSUMED_MODE_XKB):
        """Test whether a modifier is consumed by keyboard state translation
        for a key.

        Returns True if the modifier is consumed, False if it is not.
        If the modifier index is not valid in the keymap, raises
        XKBInvalidModifierIndex.
        """
        r = lib.xkb_state_mod_index_is_consumed2(self._state, key, idx, mode)
        if r == -1:
            raise XKBInvalidModifierIndex()
        return r == 1

    def mod_mask_remove_consumed(self, key, mask):
        """Remove consumed modifiers from a modifier mask for a key.

        Takes the given modifier mask, and removes all modifiers which
        are consumed for that particular key.
        """
        return lib.xkb_state_mod_mask_remove_consumed(self._state, key, mask)

    def key_get_consumed_mods(self, key,
                              mode=ConsumedMode.XKB_CONSUMED_MODE_XKB):
        """Get the mask of modifiers consumed by translating a given key.

        Returns a mask of the consumed modifiers.
        """
        return lib.xkb_state_key_get_consumed_mods2(self._state, key, mode)

    def layout_name_is_active(self, name, type):
        """Test whether a layout is active in a given keyboard state by name.

        Returns True if the layout is active, False if it is not.  If
        no layout with this name exists in the keymap, raises
        XKBLayoutDoesNotExist.

        If multiple layouts in the keymap have this name, the one with
        the lowest index is tested.
        """
        r = lib.xkb_state_layout_name_is_active(
            self._state, name.encode(), type)
        if r == -1:
            raise XKBLayoutDoesNotExist(name)
        return r == 1

    def layout_index_is_active(self, idx, type):
        """Test whether a layout is active in a given keyboard state by index.

        Returns True if the layout is active, False if it is not.  If
        the layout index is not valid in the keymap, raises
        XKBInvalidLayoutIndex.
        """
        r = lib.xkb_state_layout_index_is_active(self._state, idx, type)
        if r == -1:
            raise XKBInvalidLayoutIndex()
        return r == 1

    def led_name_is_active(self, name):
        """Test whether a LED is active in a given keyboard state by name.

        Returns True if the LED is active, False if it is not.  If no
        LED with this name exists in the keymap, raises
        XKBLEDDoesNotExist.
        """
        r = lib.xkb_state_led_name_is_active(self._state, name.encode('ascii'))
        if r == -1:
            raise XKBLEDDoesNotExist(name)
        return r == 1

    def led_index_is_active(self, idx):
        """Test whether a LED is active in a given keyboard state by index.

        Returns True if the LED is active, False if it is not.  If the
        LED index is not valid in the keymap, raises
        XKBInvalidLEDIndex.
        """
        r = lib.xkb_state_led_index_is_active(self._state, idx)
        if r == -1:
            raise XKBInvalidLEDIndex()
        return r == 1


class ComposeTable:
    """A Compose table.

    Do not instantiate this object directly.  Instead, use the various
    'compose_table_new_from_' methods of Context.
    """

    def __init__(self, context, pointer, load_method):
        self.load_method = load_method
        self._context = context

        self._table = ffi.gc(
            pointer, _keepref(lib, lib.xkb_compose_table_unref))

    # Methods to access and iterate over the compose table will be
    # added for release 1.6

    def compose_state_new(self, flags=None):
        pointer = lib.xkb_compose_state_new(self._table, flags if flags else 0)
        if not pointer:
            raise XKBComposeStateCreationFailure(
                "Couldn't create compose state")
        return ComposeState(self, pointer)


class ComposeState:
    """A Compose state object.

    The compose state maintains state for compose sequence matching,
    such as which possible sequences are being matched, and the
    position within these sequences. It acts as a simple state machine
    wherein keysyms are the input, and composed keysyms and strings
    are the output.

    The compose state is usually associated with a keyboard device.

    Do not instantiate this object directly.  Instead, use the
    'compose_state_new()' method of ComposeTable.
    """

    def __init__(self, table, pointer):
        self._table = table
        self._state = ffi.gc(
            pointer, _keepref(lib, lib.xkb_compose_state_unref))

    def feed(self, keysym):
        """Feed one keysym to the Compose sequence state machine.

        This function can advance into a compose sequence, cancel a
        sequence, start a new sequence, or do nothing in particular.
        The resulting status may be observed with get_status().

        Some keysyms, such as keysyms for modifier keys, are ignored -
        they have no effect on the status or otherwise.

        The following is a description of the possible status
        transitions, in the format CURRENT STATUS => NEXT STATUS,
        given a non-ignored input keysym `keysym`:

        NOTHING or CANCELLED or COMPOSED =>
           NOTHING   if keysym does not start a sequence.
           COMPOSING if keysym starts a sequence.
           COMPOSED  if keysym starts and terminates a single-keysym sequence.

        COMPOSING =>
           COMPOSING if keysym advances any of the currently possible
                     sequences but does not terminate any of them.
           COMPOSED  if keysym terminates one of the currently possible
                     sequences.
           CANCELLED if keysym does not advance any of the currently
                     possible sequences.

        The current Compose formats do not support multiple-keysyms.
        Therefore, if you are using a function such as
        KeyboardState.key_get_syms() and it returns more than one
        keysym, consider feeding lib.XKB_KEY_NoSymbol instead.

        A keysym param is usually obtained after a key-press event,
        with a function such as KeyboardState.key_get_one_sym().

        Returns a ComposeFeedResult indicating whether the keysym was
        ignored. This is useful, for example, if you want to keep a
        record of the sequence matched thus far.
        """
        return ComposeFeedResult(lib.xkb_compose_state_feed(
            self._state, keysym))

    def reset(self):
        """Reset the Compose sequence state machine.

        The status is set to ComposeStatus.XKB_COMPOSE_NOTHING, and
        the current sequence is discarded.
        """
        lib.xkb_compose_state_reset(self._state)

    def get_status(self):
        """Get the current status of the compose state machine.
        """
        return ComposeStatus(lib.xkb_compose_state_get_status(self._state))

    def get_utf8(self):
        """Get the result Unicode/UTF-8 string for a composed sequence.

        This function is only useful when the status is
        ComposeStatus.XKB_COMPOSE_COMPOSED.

        Returns string for composed sequence or empty string if not viable.
        """
        buffer_size = lib.xkb_compose_state_get_utf8(
            self._state, ffi.NULL, 0) + 1
        buffer = ffi.new(f"char[{buffer_size}]")
        lib.xkb_compose_state_get_utf8(self._state, buffer, buffer_size)
        return ffi.string(buffer).decode("utf8")

    def get_one_sym(self):
        """Get the result keysym for a composed sequence.

        This function is only useful when the status is
        ComposeStatus.XKB_COMPOSE_COMPOSED.

        Returns result keysym for composed sequence or
        lib.XKB_KEY_NoSymbol if not viable.
        """
        return lib.xkb_compose_state_get_one_sym(self._state)



from cffi import FFI
ffibuilder = FFI()

# Currently implemented with reference to libxkbcommon-1.5.0

ffibuilder.set_source("xkbcommon._ffi", """
#include <stdarg.h>
#include <xkbcommon/xkbcommon.h>
#include <xkbcommon/xkbcommon-keysyms.h>
#include <xkbcommon/xkbcommon-compose.h>

static void _log_handler(void *user_data,
                         enum xkb_log_level level,
                         const char *message);

static void _log_handler_internal(
    struct xkb_context *context,
    enum xkb_log_level level,
    const char *format,
    va_list args)
{
  char buf[1024];
  void *user_data;

  user_data=xkb_context_get_user_data(context);
  vsnprintf(buf, 1024, format, args);
  buf[1023]=0;
  _log_handler(user_data, level, buf);
}

void _set_log_handler_internal(struct xkb_context *context)
{
  xkb_context_set_log_fn(context, _log_handler_internal);
}

""",
                      libraries=['xkbcommon'])

ffibuilder.cdef("""
typedef ... FILE;
typedef ... va_list;

struct xkb_context;
struct xkb_keymap;
struct xkb_state;
typedef uint32_t xkb_keycode_t;
typedef uint32_t xkb_keysym_t;
typedef uint32_t xkb_layout_index_t;
typedef uint32_t xkb_layout_mask_t;
typedef uint32_t xkb_level_index_t;
typedef uint32_t xkb_mod_index_t;
typedef uint32_t xkb_mod_mask_t;
typedef uint32_t xkb_led_index_t;
typedef uint32_t xkb_led_mask_t;
#define XKB_KEYCODE_INVALID ...
#define XKB_LAYOUT_INVALID  ...
#define XKB_LEVEL_INVALID   ...
#define XKB_MOD_INVALID     ...
#define XKB_LED_INVALID     ...

#define XKB_KEYCODE_MAX     ...

#define XKB_KEY_NoSymbol    ...

struct xkb_rule_names {
    const char *rules;
    const char *model;
    const char *layout;
    const char *variant;
    const char *options;
};

int
xkb_keysym_get_name(xkb_keysym_t keysym, char *buffer, size_t size);

enum xkb_keysym_flags {
    XKB_KEYSYM_NO_FLAGS = ...,
    XKB_KEYSYM_CASE_INSENSITIVE = ...
};

xkb_keysym_t
xkb_keysym_from_name(const char *name, enum xkb_keysym_flags flags);

int
xkb_keysym_to_utf8(xkb_keysym_t keysym, char *buffer, size_t size);

uint32_t
xkb_keysym_to_utf32(xkb_keysym_t keysym);

xkb_keysym_t
xkb_keysym_to_upper(xkb_keysym_t ks);

xkb_keysym_t
xkb_keysym_to_lower(xkb_keysym_t ks);

enum xkb_context_flags {
    XKB_CONTEXT_NO_FLAGS = ...,
    XKB_CONTEXT_NO_DEFAULT_INCLUDES = ...,
    XKB_CONTEXT_NO_ENVIRONMENT_NAMES = ...,
    XKB_CONTEXT_NO_SECURE_GETENV = ...
};

struct xkb_context *
xkb_context_new(enum xkb_context_flags flags);

struct xkb_context *
xkb_context_ref(struct xkb_context *context);

void
xkb_context_unref(struct xkb_context *context);

void
xkb_context_set_user_data(struct xkb_context *context, void *user_data);

void *
xkb_context_get_user_data(struct xkb_context *context);

int
xkb_context_include_path_append(struct xkb_context *context, const char *path);

int
xkb_context_include_path_append_default(struct xkb_context *context);

int
xkb_context_include_path_reset_defaults(struct xkb_context *context);

void
xkb_context_include_path_clear(struct xkb_context *context);

unsigned int
xkb_context_num_include_paths(struct xkb_context *context);

const char *
xkb_context_include_path_get(struct xkb_context *context, unsigned int index);

enum xkb_log_level {
    XKB_LOG_LEVEL_CRITICAL = ..., /**< Log critical internal errors only. */
    XKB_LOG_LEVEL_ERROR = ...,    /**< Log all errors. */
    XKB_LOG_LEVEL_WARNING = ...,  /**< Log warnings and errors. */
    XKB_LOG_LEVEL_INFO = ...,     /**< Log information, warnings, and errors. */
    XKB_LOG_LEVEL_DEBUG = ...     /**< Log everything. */
};

void
xkb_context_set_log_level(struct xkb_context *context,
                          enum xkb_log_level level);

enum xkb_log_level
xkb_context_get_log_level(struct xkb_context *context);

void
xkb_context_set_log_verbosity(struct xkb_context *context, int verbosity);

int
xkb_context_get_log_verbosity(struct xkb_context *context);

void
xkb_context_set_log_fn(struct xkb_context *context,
                       void (*log_fn)(struct xkb_context *context,
                                      enum xkb_log_level level,
                                      const char *format, va_list args));

enum xkb_keymap_compile_flags {
    XKB_KEYMAP_COMPILE_NO_FLAGS = ...
};

struct xkb_keymap *
xkb_keymap_new_from_names(struct xkb_context *context,
                          const struct xkb_rule_names *names,
                          enum xkb_keymap_compile_flags flags);

enum xkb_keymap_format {
    XKB_KEYMAP_FORMAT_TEXT_V1 = ...
};

struct xkb_keymap *
xkb_keymap_new_from_file(struct xkb_context *context, FILE *file,
                         enum xkb_keymap_format format,
                         enum xkb_keymap_compile_flags flags);

struct xkb_keymap *
xkb_keymap_new_from_string(struct xkb_context *context, const char *string,
                           enum xkb_keymap_format format,
                           enum xkb_keymap_compile_flags flags);

struct xkb_keymap *
xkb_keymap_new_from_buffer(struct xkb_context *context, const char *buffer,
                           size_t length, enum xkb_keymap_format format,
                           enum xkb_keymap_compile_flags flags);

struct xkb_keymap *
xkb_keymap_ref(struct xkb_keymap *keymap);

void
xkb_keymap_unref(struct xkb_keymap *keymap);

#define XKB_KEYMAP_USE_ORIGINAL_FORMAT ...

char *
xkb_keymap_get_as_string(struct xkb_keymap *keymap,
                         enum xkb_keymap_format format);

xkb_keycode_t
xkb_keymap_min_keycode(struct xkb_keymap *keymap);

xkb_keycode_t
xkb_keymap_max_keycode(struct xkb_keymap *keymap);

typedef void
(*xkb_keymap_key_iter_t)(struct xkb_keymap *keymap, xkb_keycode_t key,
                         void *data);

void
xkb_keymap_key_for_each(struct xkb_keymap *keymap, xkb_keymap_key_iter_t iter,
                        void *data);

const char *
xkb_keymap_key_get_name(struct xkb_keymap *keymap, xkb_keycode_t key);

xkb_keycode_t
xkb_keymap_key_by_name(struct xkb_keymap *keymap, const char *name);

xkb_mod_index_t
xkb_keymap_num_mods(struct xkb_keymap *keymap);

const char *
xkb_keymap_mod_get_name(struct xkb_keymap *keymap, xkb_mod_index_t idx);

xkb_mod_index_t
xkb_keymap_mod_get_index(struct xkb_keymap *keymap, const char *name);

xkb_layout_index_t
xkb_keymap_num_layouts(struct xkb_keymap *keymap);

const char *
xkb_keymap_layout_get_name(struct xkb_keymap *keymap, xkb_layout_index_t idx);

xkb_layout_index_t
xkb_keymap_layout_get_index(struct xkb_keymap *keymap, const char *name);

xkb_led_index_t
xkb_keymap_num_leds(struct xkb_keymap *keymap);

const char *
xkb_keymap_led_get_name(struct xkb_keymap *keymap, xkb_led_index_t idx);

xkb_led_index_t
xkb_keymap_led_get_index(struct xkb_keymap *keymap, const char *name);

xkb_layout_index_t
xkb_keymap_num_layouts_for_key(struct xkb_keymap *keymap, xkb_keycode_t key);

xkb_level_index_t
xkb_keymap_num_levels_for_key(struct xkb_keymap *keymap, xkb_keycode_t key,
                              xkb_layout_index_t layout);

size_t
xkb_keymap_key_get_mods_for_level(struct xkb_keymap *keymap,
                                  xkb_keycode_t key,
                                  xkb_layout_index_t layout,
                                  xkb_level_index_t level,
                                  xkb_mod_mask_t *masks_out,
                                  size_t masks_size);

int
xkb_keymap_key_get_syms_by_level(struct xkb_keymap *keymap,
                                 xkb_keycode_t key,
                                 xkb_layout_index_t layout,
                                 xkb_level_index_t level,
                                 const xkb_keysym_t **syms_out);

int
xkb_keymap_key_repeats(struct xkb_keymap *keymap, xkb_keycode_t key);

struct xkb_state *
xkb_state_new(struct xkb_keymap *keymap);

struct xkb_state *
xkb_state_ref(struct xkb_state *state);

void
xkb_state_unref(struct xkb_state *state);

struct xkb_keymap *
xkb_state_get_keymap(struct xkb_state *state);

enum xkb_key_direction {
    XKB_KEY_UP = ...,
    XKB_KEY_DOWN = ...
};

enum xkb_state_component {
    XKB_STATE_MODS_DEPRESSED = ...,
    XKB_STATE_MODS_LATCHED = ...,
    XKB_STATE_MODS_LOCKED = ...,
    XKB_STATE_MODS_EFFECTIVE = ...,
    XKB_STATE_LAYOUT_DEPRESSED = ...,
    XKB_STATE_LAYOUT_LATCHED = ...,
    XKB_STATE_LAYOUT_LOCKED = ...,
    XKB_STATE_LAYOUT_EFFECTIVE = ...,
    XKB_STATE_LEDS = ...
};

enum xkb_state_component
xkb_state_update_key(struct xkb_state *state, xkb_keycode_t key,
                     enum xkb_key_direction direction);

enum xkb_state_component
xkb_state_update_mask(struct xkb_state *state,
                      xkb_mod_mask_t depressed_mods,
                      xkb_mod_mask_t latched_mods,
                      xkb_mod_mask_t locked_mods,
                      xkb_layout_index_t depressed_layout,
                      xkb_layout_index_t latched_layout,
                      xkb_layout_index_t locked_layout);

int
xkb_state_key_get_syms(struct xkb_state *state, xkb_keycode_t key,
                       const xkb_keysym_t **syms_out);

int
xkb_state_key_get_utf8(struct xkb_state *state, xkb_keycode_t key,
                       char *buffer, size_t size);

uint32_t
xkb_state_key_get_utf32(struct xkb_state *state, xkb_keycode_t key);

xkb_keysym_t
xkb_state_key_get_one_sym(struct xkb_state *state, xkb_keycode_t key);

xkb_layout_index_t
xkb_state_key_get_layout(struct xkb_state *state, xkb_keycode_t key);

xkb_level_index_t
xkb_state_key_get_level(struct xkb_state *state, xkb_keycode_t key,
                        xkb_layout_index_t layout);

enum xkb_state_match {
    XKB_STATE_MATCH_ANY = ...,
    XKB_STATE_MATCH_ALL = ...,
    XKB_STATE_MATCH_NON_EXCLUSIVE = ...
};

enum xkb_consumed_mode {
    XKB_CONSUMED_MODE_XKB = ...,
    XKB_CONSUMED_MODE_GTK = ...
};

xkb_mod_mask_t
xkb_state_serialize_mods(struct xkb_state *state,
                         enum xkb_state_component components);

xkb_layout_index_t
xkb_state_serialize_layout(struct xkb_state *state,
                           enum xkb_state_component components);

int
xkb_state_mod_name_is_active(struct xkb_state *state, const char *name,
                             enum xkb_state_component type);

int
xkb_state_mod_names_are_active(struct xkb_state *state,
                               enum xkb_state_component type,
                               enum xkb_state_match match,
                               ...);

int
xkb_state_mod_index_is_active(struct xkb_state *state, xkb_mod_index_t idx,
                              enum xkb_state_component type);

int
xkb_state_mod_indices_are_active(struct xkb_state *state,
                                 enum xkb_state_component type,
                                 enum xkb_state_match match,
                                 ...);

int
xkb_state_mod_index_is_consumed2(struct xkb_state *state, xkb_keycode_t key,
                                 xkb_mod_index_t idx,
                                 enum xkb_consumed_mode mode);

int
xkb_state_mod_index_is_consumed(struct xkb_state *state, xkb_keycode_t key,
                                xkb_mod_index_t idx);

xkb_mod_mask_t
xkb_state_mod_mask_remove_consumed(struct xkb_state *state, xkb_keycode_t key,
                                   xkb_mod_mask_t mask);

xkb_mod_mask_t
xkb_state_key_get_consumed_mods2(struct xkb_state *state, xkb_keycode_t key,
                                 enum xkb_consumed_mode mode);

xkb_mod_mask_t
xkb_state_key_get_consumed_mods(struct xkb_state *state, xkb_keycode_t key);

int
xkb_state_layout_name_is_active(struct xkb_state *state, const char *name,
                                enum xkb_state_component type);

int
xkb_state_layout_index_is_active(struct xkb_state *state,
                                 xkb_layout_index_t idx,
                                 enum xkb_state_component type);

int
xkb_state_led_name_is_active(struct xkb_state *state, const char *name);

int
xkb_state_led_index_is_active(struct xkb_state *state, xkb_led_index_t idx);

struct xkb_compose_table;
struct xkb_compose_state;
enum xkb_compose_compile_flags {
    XKB_COMPOSE_COMPILE_NO_FLAGS = ...
};
enum xkb_compose_format {
    XKB_COMPOSE_FORMAT_TEXT_V1 = ...
};

struct xkb_compose_table *
xkb_compose_table_new_from_locale(struct xkb_context *context,
                                  const char *locale,
                                  enum xkb_compose_compile_flags flags);

struct xkb_compose_table *
xkb_compose_table_new_from_file(struct xkb_context *context,
                                FILE *file,
                                const char *locale,
                                enum xkb_compose_format format,
                                enum xkb_compose_compile_flags flags);

struct xkb_compose_table *
xkb_compose_table_new_from_buffer(struct xkb_context *context,
                                  const char *buffer, size_t length,
                                  const char *locale,
                                  enum xkb_compose_format format,
                                  enum xkb_compose_compile_flags flags);

struct xkb_compose_table *
xkb_compose_table_ref(struct xkb_compose_table *table);

void
xkb_compose_table_unref(struct xkb_compose_table *table);

enum xkb_compose_state_flags {
    XKB_COMPOSE_STATE_NO_FLAGS = ...
};

struct xkb_compose_state *
xkb_compose_state_new(struct xkb_compose_table *table,
                      enum xkb_compose_state_flags flags);

struct xkb_compose_state *
xkb_compose_state_ref(struct xkb_compose_state *state);

void
xkb_compose_state_unref(struct xkb_compose_state *state);

struct xkb_compose_table *
xkb_compose_state_get_compose_table(struct xkb_compose_state *state);

enum xkb_compose_status {
    XKB_COMPOSE_NOTHING = ...,
    XKB_COMPOSE_COMPOSING = ...,
    XKB_COMPOSE_COMPOSED = ...,
    XKB_COMPOSE_CANCELLED = ...
};

enum xkb_compose_feed_result {
    XKB_COMPOSE_FEED_IGNORED = ...,
    XKB_COMPOSE_FEED_ACCEPTED = ...
};

enum xkb_compose_feed_result
xkb_compose_state_feed(struct xkb_compose_state *state,
                       xkb_keysym_t keysym);

void
xkb_compose_state_reset(struct xkb_compose_state *state);

enum xkb_compose_status
xkb_compose_state_get_status(struct xkb_compose_state *state);

int
xkb_compose_state_get_utf8(struct xkb_compose_state *state,
                           char *buffer, size_t size);

xkb_keysym_t
xkb_compose_state_get_one_sym(struct xkb_compose_state *state);

void _set_log_handler_internal(struct xkb_context *context);

extern "Python" void _log_handler(void *user_data,
                                  enum xkb_log_level level,
                                  const char *message);

extern "Python" void _key_for_each_helper(
    struct xkb_keymap *keymap,
    xkb_keycode_t key,
    void *data);

void free(void *ptr);

""")

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



# NB these tests are intended to check whether the python bindings are
# working, not whether libxkbcommon itself is working!

from unittest import TestCase

from xkbcommon import xkb

from tests.data import sample_keymap_string, sample_keymap_bytes, \
    sample_compose_bytes

import os
import tempfile
from io import BytesIO
import array

# A directory that is guaranteed to exist
testdir = os.path.dirname(os.path.abspath(__file__))
# A path that is guaranteed not to exist
nonexistent = os.path.join(testdir, "must-not-exist")
# A file that is guaranteed to exist
testfile = os.path.abspath(__file__)


class TestKeysym(TestCase):
    def test_keysym_get_name(self):
        self.assertEqual(xkb.keysym_get_name(0xff68), "Find")

    def test_keysym_from_name(self):
        self.assertEqual(xkb.keysym_from_name("space"), 0x20)

    def test_keysym_to_string(self):
        self.assertEqual(xkb.keysym_to_string(0x20), ' ')

    def test_keysym_to_upper(self):
        # Lower case to upper case A
        self.assertEqual(xkb.keysym_to_upper(0x61), 0x41)

    def test_keysym_to_lower(self):
        # Upper case to lower case A
        self.assertEqual(xkb.keysym_to_lower(0x41), 0x61)


class TestContext(TestCase):
    @classmethod
    def setUpClass(cls):
        # Create temporary files to use while testing.
        cls._sample_keymap_file = tempfile.NamedTemporaryFile(
            mode='w+b')
        cls._sample_keymap_file.write(sample_keymap_bytes)
        cls._sample_compose_file = tempfile.NamedTemporaryFile(
            mode='w+b')
        cls._sample_compose_file.write(sample_compose_bytes)

    @classmethod
    def tearDownClass(cls):
        cls._sample_keymap_file.close()
        cls._sample_compose_file.close()

    def test_create(self):
        xkb.Context()

    def test_no_secure_getenv(self):
        # We are unlikely to be testing in an environment where
        # secure_getenv() will return NULL instead of a value string,
        # so just check that we are able to create a context with this
        # flag set
        xkb.Context(no_secure_getenv=True)

    def test_default_includes(self):
        ctx = xkb.Context(no_default_includes=True)
        self.assertEqual(len(list(ctx.include_path())), 0)

    def test_include_path_append_exists(self):
        ctx = xkb.Context()
        ctx.include_path_append(testdir)
        self.assertIn(testdir, ctx.include_path())

    def test_include_path_append_file(self):
        ctx = xkb.Context()
        with self.assertRaises(xkb.XKBPathError):
            ctx.include_path_append(testfile)
        self.assertNotIn(testfile, ctx.include_path())

    def test_include_path_append_does_not_exist(self):
        ctx = xkb.Context()
        with self.assertRaises(xkb.XKBPathError):
            ctx.include_path_append(nonexistent)
        self.assertNotIn(nonexistent, ctx.include_path())

    def test_num_include_paths(self):
        ctx = xkb.Context()
        default_num_include_paths = ctx.num_include_paths()
        ctx.include_path_append(testdir)
        self.assertEqual(ctx.num_include_paths(),
                         default_num_include_paths + 1)

    def test_set_log_level(self):
        ctx = xkb.Context()
        self.assertEqual(ctx.get_log_level(), xkb.lib.XKB_LOG_LEVEL_ERROR)
        ctx.set_log_level(xkb.lib.XKB_LOG_LEVEL_DEBUG)
        self.assertEqual(ctx.get_log_level(), xkb.lib.XKB_LOG_LEVEL_DEBUG)

    def test_set_log_verbosity(self):
        ctx = xkb.Context()
        self.assertEqual(ctx.get_log_verbosity(), 0)
        ctx.set_log_verbosity(10)
        self.assertEqual(ctx.get_log_verbosity(), 10)

    def test_set_log_handler(self):
        messages = []

        def handler(context, level, message):
            messages.append(message)

        ctx = xkb.Context()
        ctx.set_log_level(xkb.lib.XKB_LOG_LEVEL_DEBUG)
        ctx.set_log_verbosity(10)
        ctx.set_log_fn(handler)
        ctx.keymap_new_from_string(sample_keymap_string)
        self.assertNotEqual(len(messages), 0)

    def test_keymap_new_from_names_with_args(self):
        # NB this test requires that suitable keymaps are installed
        # wherever xkbcommon expects to find them
        ctx = xkb.Context(no_environment_names=True)
        km = ctx.keymap_new_from_names(
            rules="evdev", model="pc105", layout="gb",
            variant="dvorak", options="terminate:ctrl_alt_bksp")
        self.assertIsNotNone(km)

    def test_keymap_new_from_names(self):
        ctx = xkb.Context()
        km = ctx.keymap_new_from_names()
        self.assertIsNotNone(km)

    def test_keymap_new_from_file_mmap(self):
        ctx = xkb.Context()
        self._sample_keymap_file.seek(0)
        km = ctx.keymap_new_from_file(self._sample_keymap_file)
        self.assertIsNotNone(km, None)
        self.assertEqual(km.load_method, "mmap_file")

    def test_keymap_new_from_file_no_mmap(self):
        ctx = xkb.Context()
        memfile = BytesIO(sample_keymap_bytes)
        km = ctx.keymap_new_from_file(memfile)
        self.assertIsNotNone(km)
        self.assertEqual(km.load_method, "read_file")

    def test_keymap_new_from_string(self):
        ctx = xkb.Context()
        km = ctx.keymap_new_from_string(sample_keymap_string)
        self.assertIsNotNone(km, None)
        self.assertEqual(km.load_method, "string")

    def test_keymap_new_from_buffer(self):
        ctx = xkb.Context()
        test_data = array.array('b', sample_keymap_bytes)
        km = ctx.keymap_new_from_buffer(test_data)
        self.assertIsNotNone(km)
        self.assertEqual(km.load_method, "buffer")
        test_data.extend([0] * 10)
        length = len(sample_keymap_bytes)
        km = ctx.keymap_new_from_buffer(test_data, length=length)
        self.assertIsNotNone(km)

    def test_compose_table_new_from_locale(self):
        ctx = xkb.Context()
        ct = ctx.compose_table_new_from_locale("en_GB.UTF-8")
        self.assertIsNotNone(ct)
        self.assertEqual(ct.load_method, "locale")

    def test_compose_table_new_from_locale_fail(self):
        ctx = xkb.Context()
        with self.assertRaises(xkb.XKBComposeTableCreationFailure):
            ctx.compose_table_new_from_locale("not-a-locale")

    def test_compose_table_new_from_file_mmap(self):
        ctx = xkb.Context()
        self._sample_compose_file.seek(0)
        ct = ctx.compose_table_new_from_file(
            self._sample_compose_file, "en_US.UTF-8")
        self.assertIsNotNone(ct)
        self.assertEqual(ct.load_method, "mmap_file")

    def test_compose_table_new_from_file_no_mmap(self):
        ctx = xkb.Context()
        memfile = BytesIO(sample_compose_bytes)
        ct = ctx.compose_table_new_from_file(memfile, "en_US.UTF-8")
        self.assertIsNotNone(ct)
        self.assertEqual(ct.load_method, "read_file")

    def test_compose_table_new_from_buffer(self):
        ctx = xkb.Context()
        test_data = sample_compose_bytes
        ct = ctx.compose_table_new_from_buffer(test_data, "en_US.UTF-8")
        self.assertIsNotNone(ct)
        self.assertEqual(ct.load_method, "buffer")


# This class makes use of the details of the sample keymap in
# sample_keymap_string.
class TestKeymap(TestCase):
    @classmethod
    def setUpClass(cls):
        cls._ctx = xkb.Context()
        cls.km = cls._ctx.keymap_new_from_string(sample_keymap_string)

    def test_keymap_get_as_string(self):
        kms = self.km.get_as_string()
        self.assertNotEqual(len(kms), 0)

    def test_keymap_min_keycode(self):
        self.assertEqual(self.km.min_keycode(), 9)

    def test_keymap_max_keycode(self):
        self.assertEqual(self.km.max_keycode(), 253)

    def test_keymap_iterate_over_keycodes(self):
        self.assertEqual(len(list(self.km)), 245)

    def test_keymap_key_names_roundtrip(self):
        # Test both key_get_name() and key_by_name()
        for keycode in iter(self.km):
            try:
                name = self.km.key_get_name(keycode)
                self.assertEqual(self.km.key_by_name(name), keycode)
            except xkb.XKBInvalidKeycode:
                pass

    def test_keymap_key_get_name_invalid(self):
        # Keycode 93 does not exist in our sample keymap
        with self.assertRaises(xkb.XKBInvalidKeycode):
            self.km.key_get_name(93)

    def test_keymap_key_by_name_invalid(self):
        # Keycode 'WIBL' does not exist in our sample keymap
        with self.assertRaises(xkb.XKBKeyDoesNotExist):
            self.km.key_by_name("WIBL")

    def test_keymap_num_mods(self):
        self.assertEqual(self.km.num_mods(), 21)

    def test_keymap_mod_get_name(self):
        self.assertEqual(self.km.mod_get_name(0), "Shift")

    def test_keymap_mod_get_name_fail(self):
        with self.assertRaises(xkb.XKBInvalidModifierIndex):
            self.km.mod_get_name(self.km.num_mods())

    def test_keymap_mod_get_index(self):
        self.assertEqual(self.km.mod_get_index("Shift"), 0)

    def test_keymap_mod_get_index_fail(self):
        with self.assertRaises(xkb.XKBModifierDoesNotExist):
            self.km.mod_get_index("wibble")

    def test_keymap_num_layouts(self):
        self.assertEqual(self.km.num_layouts(), 2)

    def test_keymap_layout_get_name(self):
        self.assertEqual(self.km.layout_get_name(0), "English (UK)")

    def test_keymap_layout_get_name_fail(self):
        with self.assertRaises(xkb.XKBInvalidLayoutIndex):
            self.km.layout_get_name(self.km.num_layouts())

    def test_keymap_layout_get_index(self):
        self.assertEqual(self.km.layout_get_index("English (UK)"), 0)

    def test_keymap_layout_get_index_fail(self):
        with self.assertRaises(xkb.XKBLayoutDoesNotExist):
            self.km.layout_get_index("wibble")

    def test_keymap_num_leds(self):
        self.assertEqual(self.km.num_leds(), 14)

    def test_keymap_led_get_name(self):
        self.assertEqual(self.km.led_get_name(0), "Caps Lock")

    def test_keymap_led_get_name_fail(self):
        with self.assertRaises(xkb.XKBInvalidLEDIndex):
            self.km.led_get_name(self.km.num_leds())

    def test_keymap_led_get_index(self):
        self.assertEqual(self.km.led_get_index("Caps Lock"), 0)

    def test_keymap_led_get_index_fail(self):
        with self.assertRaises(xkb.XKBLEDDoesNotExist):
            self.km.led_get_index("wibble")

    def test_keymap_num_layouts_for_key(self):
        key = next(iter(self.km))
        self.assertEqual(self.km.num_layouts_for_key(key), 1)

    def test_keymap_num_levels_for_key(self):
        key = next(iter(self.km))
        self.assertEqual(self.km.num_levels_for_key(key, 0), 1)

    def test_keymap_key_get_mods_for_level(self):
        first_invalid_mask = 1 << self.km.num_mods()
        for key in self.km:
            for layout in range(self.km.num_layouts_for_key(key)):
                for level in range(self.km.num_levels_for_key(key, layout)):
                    r = self.km.key_get_mods_for_level(key, layout, level)
                    for mm in r:
                        self.assertLess(mm, first_invalid_mask)

    def test_keymap_key_get_syms_by_level(self):
        for key in self.km:
            r = self.km.key_get_syms_by_level(key, 0, 0)
            for k in r:
                # If we are returned an invalid keysym, this will raise
                # XKBInvalidKeysym
                self.assertTrue(xkb.keysym_get_name(k))

    def test_keymap_key_repeats(self):
        key = next(iter(self.km))
        self.assertEqual(self.km.key_repeats(key), True)

    def test_keymap_state_new(self):
        self.assertIsNotNone(self.km.state_new())


class TestBitEnum(TestCase):
    def test_StateComponent(self):
        a = xkb.XKB_STATE_MODS_DEPRESSED
        b = xkb.XKB_STATE_MODS_LATCHED
        self.assertIsInstance(a | b, xkb.StateComponent)

        # The order in which enum members are listed is implementation
        # defined and does in practice vary between different Python
        # versions. Both possibilities should be acceptable.

        self.assertIn(
            str(a | b),
            ("StateComponent.XKB_STATE_MODS_LATCHED|XKB_STATE_MODS_DEPRESSED",
             "StateComponent.XKB_STATE_MODS_DEPRESSED|XKB_STATE_MODS_LATCHED"))
        c = a | 0x1000000
        self.assertIn(
            str(c),
            ("StateComponent.16777216|XKB_STATE_MODS_DEPRESSED",
             "StateComponent.XKB_STATE_MODS_DEPRESSED|16777216"))

    def test_StateMatch(self):
        self.assertIn(
            str(xkb.XKB_STATE_MATCH_ANY | xkb.XKB_STATE_MATCH_ALL),
            ("StateMatch.XKB_STATE_MATCH_ALL|XKB_STATE_MATCH_ANY",
             "StateMatch.XKB_STATE_MATCH_ANY|XKB_STATE_MATCH_ALL"))


class TestKeyboardState(TestCase):
    capslock = 66
    space = 65

    @classmethod
    def setUpClass(cls):
        cls._ctx = xkb.Context()
        cls.km = cls._ctx.keymap_new_from_string(sample_keymap_string)
        cls.lock = cls.km.mod_get_index("Lock")
        cls.numlock = cls.km.mod_get_index("NumLock")
        cls.badmod = cls.km.num_mods()

    def test_state_get_keymap(self):
        state = self.km.state_new()
        self.assertEqual(state.get_keymap(), self.km)
        self.assertEqual(state.keymap, self.km)

    def test_state_update_key(self):
        state = self.km.state_new()
        self.assertEqual(
            state.mod_name_is_active("Lock", xkb.XKB_STATE_MODS_LOCKED),
            False)
        keydown = state.update_key(self.capslock, xkb.XKB_KEY_DOWN)
        self.assertEqual(
            keydown, xkb.XKB_STATE_LEDS | xkb.XKB_STATE_MODS_EFFECTIVE
            | xkb.XKB_STATE_MODS_LOCKED | xkb.XKB_STATE_MODS_DEPRESSED)
        self.assertIsInstance(keydown, xkb.StateComponent)
        self.assertIsInstance(keydown, int)
        keyup = state.update_key(self.capslock, xkb.XKB_KEY_UP)
        self.assertEqual(
            keyup, xkb.XKB_STATE_MODS_DEPRESSED)
        self.assertEqual(
            state.mod_name_is_active("Lock", xkb.XKB_STATE_MODS_LOCKED),
            True)

    def test_state_update_mask(self):
        master_state = self.km.state_new()
        slave_state = self.km.state_new()
        master_state.update_key(self.capslock, xkb.XKB_KEY_DOWN)
        master_state.update_key(self.capslock, xkb.XKB_KEY_UP)
        depressed_mods = master_state.serialize_mods(
            xkb.XKB_STATE_MODS_DEPRESSED)
        latched_mods = master_state.serialize_mods(
            xkb.XKB_STATE_MODS_LATCHED)
        locked_mods = master_state.serialize_mods(
            xkb.XKB_STATE_MODS_LOCKED)
        depressed_layout = master_state.serialize_layout(
            xkb.XKB_STATE_LAYOUT_DEPRESSED)
        latched_layout = master_state.serialize_layout(
            xkb.XKB_STATE_LAYOUT_LATCHED)
        locked_layout = master_state.serialize_layout(
            xkb.XKB_STATE_LAYOUT_LOCKED)
        self.assertEqual(
            slave_state.mod_name_is_active("Lock", xkb.XKB_STATE_MODS_LOCKED),
            False)
        r = slave_state.update_mask(
            depressed_mods, latched_mods, locked_mods,
            depressed_layout, latched_layout, locked_layout)
        self.assertIsInstance(r, int)
        self.assertIsInstance(r, xkb.StateComponent)
        self.assertEqual(
            r, xkb.XKB_STATE_LEDS | xkb.XKB_STATE_MODS_EFFECTIVE
            | xkb.XKB_STATE_MODS_LOCKED)
        self.assertEqual(
            slave_state.mod_name_is_active("Lock", xkb.XKB_STATE_MODS_LOCKED),
            True)

    def test_state_key_get_syms(self):
        state = self.km.state_new()
        syms = state.key_get_syms(self.space)
        self.assertEqual(syms, [32])

    def test_state_ket_get_string(self):
        state = self.km.state_new()
        self.assertEqual(state.key_get_string(self.capslock), "")
        self.assertEqual(state.key_get_string(self.space), " ")

    def test_state_key_get_one_sym(self):
        state = self.km.state_new()
        self.assertEqual(state.key_get_one_sym(self.capslock), 65509)
        self.assertEqual(state.key_get_one_sym(self.space), 32)

    def test_state_key_get_layout(self):
        state = self.km.state_new()
        self.assertEqual(state.key_get_layout(self.space), 0)

    def test_state_key_get_layout_fail(self):
        state = self.km.state_new()
        with self.assertRaises(xkb.XKBInvalidLayoutIndex):
            state.key_get_layout(0)

    def test_state_key_get_level(self):
        state = self.km.state_new()
        self.assertEqual(state.key_get_level(self.space, 0), 0)

    def test_state_key_get_level_fail(self):
        state = self.km.state_new()
        with self.assertRaises(xkb.XKBInvalidLayoutIndex):
            state.key_get_level(0, 0)

    # serialize_mods and serialize_layout are tested in test_state_update_mask

    # mod_name_is_active is tested in test_state_update_key

    def test_state_mod_names_are_active(self):
        state = self.km.state_new()
        self.assertFalse(
            state.mod_names_are_active(xkb.XKB_STATE_MODS_LOCKED,
                                       xkb.XKB_STATE_MATCH_ANY,
                                       ["Lock", "NumLock"]))
        state.update_key(self.capslock, xkb.XKB_KEY_DOWN)
        state.update_key(self.capslock, xkb.XKB_KEY_UP)
        self.assertTrue(
            state.mod_names_are_active(xkb.XKB_STATE_MODS_LOCKED,
                                       xkb.XKB_STATE_MATCH_ANY,
                                       ["Lock", "NumLock"]))

    def test_state_mod_index_is_active(self):
        state = self.km.state_new()
        self.assertFalse(
            state.mod_index_is_active(self.lock, xkb.XKB_STATE_MODS_LOCKED))
        state.update_key(self.capslock, xkb.XKB_KEY_DOWN)
        state.update_key(self.capslock, xkb.XKB_KEY_UP)
        self.assertTrue(
            state.mod_index_is_active(self.lock, xkb.XKB_STATE_MODS_LOCKED))

    def test_state_mod_index_is_active_fail(self):
        state = self.km.state_new()
        with self.assertRaises(xkb.XKBInvalidModifierIndex):
            state.mod_index_is_active(self.badmod,
                                      xkb.XKB_STATE_MODS_LOCKED)

    def test_state_mod_indices_are_active(self):
        state = self.km.state_new()
        self.assertFalse(
            state.mod_indices_are_active(xkb.XKB_STATE_MODS_LOCKED,
                                         xkb.XKB_STATE_MATCH_ANY,
                                         [self.lock, self.numlock]))
        state.update_key(self.capslock, xkb.XKB_KEY_DOWN)
        state.update_key(self.capslock, xkb.XKB_KEY_UP)
        self.assertTrue(
            state.mod_indices_are_active(xkb.XKB_STATE_MODS_LOCKED,
                                         xkb.XKB_STATE_MATCH_ANY,
                                         [self.lock, self.numlock]))

    def test_state_mod_indices_are_active_fail(self):
        state = self.km.state_new()
        with self.assertRaises(xkb.XKBInvalidModifierIndex):
            state.mod_indices_are_active(xkb.XKB_STATE_MODS_LOCKED,
                                         xkb.XKB_STATE_MATCH_ANY,
                                         [self.badmod])

    def test_state_mod_index_is_consumed(self):
        state = self.km.state_new()
        self.assertIsInstance(
            state.mod_index_is_consumed(self.space, self.lock),
            bool)

    def test_state_mod_index_is_consumed_fail(self):
        state = self.km.state_new()
        with self.assertRaises(xkb.XKBInvalidModifierIndex):
            state.mod_index_is_consumed(self.space, self.badmod)

    def test_state_mod_mask_remove_consumed(self):
        state = self.km.state_new()
        self.assertEqual(
            state.mod_mask_remove_consumed(self.space, 0), 0)

    def test_state_key_get_consumed_mods(self):
        state = self.km.state_new()
        self.assertEqual(
            state.key_get_consumed_mods(self.space), 0)

    def test_state_layout_name_is_active(self):
        state = self.km.state_new()
        self.assertTrue(
            state.layout_name_is_active("English (UK)",
                                        xkb.XKB_STATE_LAYOUT_EFFECTIVE))

    def test_state_layout_name_is_active_fail(self):
        state = self.km.state_new()
        with self.assertRaises(xkb.XKBLayoutDoesNotExist):
            state.layout_name_is_active("wibble",
                                        xkb.XKB_STATE_LAYOUT_EFFECTIVE)

    def test_state_layout_index_is_active(self):
        state = self.km.state_new()
        self.assertTrue(
            state.layout_index_is_active(0, xkb.XKB_STATE_LAYOUT_EFFECTIVE))

    def test_state_layout_index_is_active_fail(self):
        state = self.km.state_new()
        with self.assertRaises(xkb.XKBInvalidLayoutIndex):
            state.layout_index_is_active(self.km.num_layouts(),
                                         xkb.XKB_STATE_LAYOUT_EFFECTIVE)

    def test_state_led_name_is_active(self):
        state = self.km.state_new()
        self.assertFalse(state.led_name_is_active("Caps Lock"))
        state.update_key(self.capslock, xkb.XKB_KEY_DOWN)
        state.update_key(self.capslock, xkb.XKB_KEY_UP)
        self.assertTrue(state.led_name_is_active("Caps Lock"))

    def test_state_led_name_is_active_fail(self):
        state = self.km.state_new()
        with self.assertRaises(xkb.XKBLEDDoesNotExist):
            state.led_name_is_active("wibble")

    def test_state_led_index_is_active(self):
        state = self.km.state_new()
        capslockled = self.km.led_get_index("Caps Lock")
        self.assertFalse(state.led_index_is_active(capslockled))
        state.update_key(self.capslock, xkb.XKB_KEY_DOWN)
        state.update_key(self.capslock, xkb.XKB_KEY_UP)
        self.assertTrue(state.led_index_is_active(capslockled))

    def test_state_led_index_is_active_fail(self):
        state = self.km.state_new()
        with self.assertRaises(xkb.XKBInvalidLEDIndex):
            state.led_index_is_active(self.km.num_leds())


class TestComposeTable(TestCase):
    @classmethod
    def setUpClass(cls):
        ctx = xkb.Context()
        cls.ct = ctx.compose_table_new_from_buffer(
            sample_compose_bytes, "en_US.UTF-8")

    def test_compose_table_create_compose_state(self):
        state = self.ct.compose_state_new()
        self.assertIsInstance(state, xkb.ComposeState)


class TestCompose(TestCase):
    XKB_KEYSYM_OHORNTILDE = 0x1001ee0
    UTF8_OHORNTILDE = "Ỡ"

    XKB_KEYSYM_DEAD_CIRCUMFLEX = 0xfe52
    UTF_CIRCUMFLEX = "^"

    XKB_KEYSYM_DEAD_TILDE = 0xfe53
    XKB_KEYSYM_DEAD_HORN = 0xfe62
    XKB_KEYSYM_O = 0x004f
    XKB_KEYSYM_F = 0x0066

    @classmethod
    def setUpClass(cls):
        ctx = xkb.Context()
        ct = ctx.compose_table_new_from_buffer(
            sample_compose_bytes, "en_US.UTF-8")
        cls.compose = ct.compose_state_new()

    def test_compose_initial_status(self):
        self.compose.reset()
        self.assertEqual(
            self.compose.get_status(),
            xkb.ComposeStatus.XKB_COMPOSE_NOTHING
        )

    def test_compose_feed_double_dead_circumflex(self):
        self.compose.reset()

        self.assertEqual(
            self.compose.feed(self.XKB_KEYSYM_DEAD_CIRCUMFLEX),
            xkb.ComposeFeedResult.XKB_COMPOSE_FEED_ACCEPTED
        )
        self.assertEqual(
            self.compose.get_status(),
            xkb.ComposeStatus.XKB_COMPOSE_COMPOSING
        )

        self.assertEqual(
            self.compose.feed(self.XKB_KEYSYM_DEAD_CIRCUMFLEX),
            xkb.ComposeFeedResult.XKB_COMPOSE_FEED_ACCEPTED
        )

        self.assertEqual(
            self.compose.get_status(),
            xkb.ComposeStatus.XKB_COMPOSE_COMPOSED
        )
        self.assertEqual(self.compose.get_utf8(), self.UTF_CIRCUMFLEX)

    def test_compose_feed_non_composing(self):
        self.compose.reset()

        self.assertEqual(
            self.compose.feed(self.XKB_KEYSYM_DEAD_TILDE),
            xkb.ComposeFeedResult.XKB_COMPOSE_FEED_ACCEPTED
        )
        self.assertEqual(
            self.compose.feed(self.XKB_KEYSYM_F),
            xkb.ComposeFeedResult.XKB_COMPOSE_FEED_ACCEPTED
        )
        self.assertEqual(
            self.compose.get_status(), xkb.ComposeStatus.XKB_COMPOSE_CANCELLED
        )

        self.assertEqual(
            self.compose.feed(self.XKB_KEYSYM_F),
            xkb.ComposeFeedResult.XKB_COMPOSE_FEED_ACCEPTED
        )
        self.assertEqual(
            self.compose.get_status(), xkb.ComposeStatus.XKB_COMPOSE_NOTHING
        )

    def test_compose_feed_multi_composing(self):
        self.compose.reset()

        self.assertEqual(
            self.compose.feed(self.XKB_KEYSYM_DEAD_TILDE),
            xkb.ComposeFeedResult.XKB_COMPOSE_FEED_ACCEPTED
        )
        self.assertEqual(
            self.compose.feed(self.XKB_KEYSYM_DEAD_HORN),
            xkb.ComposeFeedResult.XKB_COMPOSE_FEED_ACCEPTED
        )
        self.assertEqual(
            self.compose.feed(self.XKB_KEYSYM_O),
            xkb.ComposeFeedResult.XKB_COMPOSE_FEED_ACCEPTED
        )
        self.assertEqual(
            self.compose.get_status(), xkb.ComposeStatus.XKB_COMPOSE_COMPOSED
        )
        self.assertEqual(self.compose.get_utf8(), self.UTF8_OHORNTILDE)
        self.assertEqual(self.compose.get_one_sym(), self.XKB_KEYSYM_OHORNTILDE)



# coding: utf-8
"""
    xkbcommon
    ~~~~~~~~~

    CFFI-based libxkbcommon bindings for Python.  See README.rst for details.

    :copyright: Copyright 2016 by Stephen Early
    :license: MIT, see LICENSE for details.
"""
